import torch
from torch import nn

class BaseModel(nn.Module):
    def __init__(self, num_features: int, **kwargs):
        super().__init__(**kwargs)
        self.num_features = num_features

    def forward(self, x):
        raise NotImplementedError

    def train(self, data):
        raise NotImplementedError

    def infer(self, data):
        raise NotImplementedError

    def save_model(self, path):
        torch.save(self.state_dict(), path)

    def load_model(self, path):
        self.load_state_dict(torch.load(path))



from abc import ABC, abstractmethod
from pathlib import Path
# from utils import *
from scapy.all import *
import pickle

import json
from io import TextIOWrapper

class LazyInitializationMixin:
    def lazy_init(self, **kwargs):
        
        for k, v in kwargs.items():
            if k in self.allowed:
                setattr(self, k, v)
            else:
                raise ValueError(f"{k} not allowed")
            setattr(self, k, v)
            self.allowed.remove(k)

    def start(self, **kwargs):
        assigned=list(self.allowed)
        for k, v in kwargs.items():
            if k in self.allowed:
                setattr(self, k, v)
            else:
                raise ValueError(f"{k} not allowed")
            assigned.remove(k)

        if len(assigned)>0:
            raise ValueError("Must assign the following variables",",".join(assigned))

        return self.entry()

    def __rrshift__(self, other):
        return self.start(**other)

class JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Path):
            return str(obj)
        if isinstance(obj, TextIOWrapper):
            return obj.name
        if isinstance(obj, np.float32):
            return float(obj)
        return super().default(obj)

def load_dataset_info():
    with open("./datasets/data_info.json", "r") as f:
        data_info = json.load(f)
    return data_info

def save_dataset_info(data_info):
    with open("./datasets/data_info.json", "w") as f:
        json.dump(data_info, f, indent=4, cls=JSONEncoder)

class BaseTrafficFeatureExtractor(ABC, LazyInitializationMixin):
    def __init__(self, **kwargs):
        """base feature extractor. By default three variables must be
        set: dataset_name, file_name, and state.
        """
        self.allowed = ["dataset_name", "file_name", "state"]
        self.lazy_init(**kwargs)
        self.entry = self.extract_features

    @abstractmethod
    def update(self, traffic_vector):
        """Updates the feature extractor with traffic_vector, and returns
        the features

        Args:
            traffic_vector (array): traffic vector extracted from the packets
        """
        pass

    @abstractmethod
    def peek(self, traffic_vectors):
        """applies fake update to the feature extractor, does not actually
        update the state of feature extractor. Not required but used for LM attack.
        returns a list of features corresonding to the traffic vectors

        Args:
            traffic_vectors (list of array): list of traffic vectors to be updated
        """
        pass

    @abstractmethod
    def get_traffic_vector(self, packet):
        """extracts traffic vectors from the raw packet,
        returns the extracted traffic vector.

        Args:
            packet (scapy packet): input packet
        """
        pass

    def setup(self):
        """set up the feature extractor. By default it opens the input pcap
        file, the output feature csv file, output meta data (AKA traffic vector) csv file,
        and sets the state of the extractor.
        It also initializes count (number of packets processed), skipped (number of
        packets skipped), and written (number of features extracted)
        """
        self.path = Path(
            f"../../datasets/{self.dataset_name}/pcap/{self.file_name}.pcap"
        )

        feature_file = Path(
            f"../../datasets/{self.dataset_name}/{self.name}/{self.file_name}.csv"
        )
        feature_file.parent.mkdir(parents=True, exist_ok=True)
        meta_file = Path(
            f"../../datasets/{self.dataset_name}/{self.name}/{self.file_name}_meta.csv"
        )
        meta_file.parent.mkdir(parents=True, exist_ok=True)

        self.feature_file = open(feature_file, "w")
        self.meta_file = open(meta_file, "w")
        self.feature_file.write(",".join(self.get_headers()) + "\n")
        self.meta_file.write(",".join(self.get_meta_headers()) + "\n")

        self.count = 0
        self.skipped = 0


        self.input_pcap = PcapReader(str(self.path))

        if self.state is not None:
            self.reset_state = False
            self.save_state = False
            self.offset_timestamp = True
        else:
            self.reset_state = True
            self.save_state = True
            self.offset_timestamp = False
        self.offset_time = None

    @abstractmethod
    def get_headers(self):
        """returns the names of the features"""
        pass

    @abstractmethod
    def get_meta_headers(self):
        """returns the names of the traffic vectors"""
        pass

    def teardown(self):
        """closes the opened files, write the files in the dataset_info json
        file, save the state.
        """
        self.meta_file.close()
        self.feature_file.close()
        self.input_pcap.close()
        # save file information
        data_info = load_dataset_info()

        if self.dataset_name not in data_info.keys():
            data_info[self.dataset_name] = {}

        if self.name not in data_info[self.dataset_name].keys():
            data_info[self.dataset_name][self.name] = {}

        data_info[self.dataset_name][self.name][self.file_name] = {
            "pcap_path": self.path,
            "feature_path": self.feature_file,
            "meta_path": self.meta_file,
            "num_rows": self.count,
        }

        save_dataset_info(data_info)
        print(
            f"skipped: {self.skipped} processed: {self.count+self.skipped} written: {self.count}"
        )

        if self.save_state:
            state_path = Path(
                f"../../datasets/{self.dataset_name}/{self.name}/state.pkl"
            )
            state_path.parent.mkdir(parents=True, exist_ok=True)
            with open(state_path, "wb") as pf:
                pickle.dump(self.state, pf)

    @abstractmethod
    def extract_features(self):
        """The main entry point of feature extractor, this function
        should define the process of extracting the features from input
        pcap file
        """
        pass



import torch
import pickle
import numpy as np

from scipy.stats import norm
from scipy.cluster.hierarchy import linkage, to_tree

from titli.ids.torch_kitnet import _TorchKitNET # type: ignore

class KitNET:
    """    
        n                   : the number of features in your input dataset (i.e., x \in R^n)
        max_autoencoder_size: the maximum size of any autoencoder in the ensemble layer
        FM_grace_period     : the number of instances which will be taken to learn the feature mapping.
                              If 'None', then FM_grace_period=AM_grace_period
        AD_grace_period     : the number of instances the network will learn from before producing
                              anomaly scores
        learning_rate       : stochastic gradient descent learning rate for all autoencoders.
        hidden_ratio        : the default ratio of hidden to visible neurons.
        feature_map         : One may optionally provide a feature map instead of learning one. The map must be
                              a list, where the i-th entry contains a list of the feature indices to be assingned
                              to the i-th autoencoder in the ensemble. For example, [[2,5,3],[4,0,1],[6,7]]
        normalize           : boolean, whether to 0-1 normalize the input data (default: True)
        input_precision     : integer, number of significant figures in the input data
        quantize            : boolean, whether to quantize the input data (default: None)
        model_path          : path to save the model
    """
    def __init__(self, n,
                 max_autoencoder_size=10,
                 FM_grace_period=None,
                 AD_grace_period=10000,
                 learning_rate=0.1,
                 hidden_ratio=0.75,
                 feature_map=None,
                 normalize=True,
                 input_precision=None,
                 quantize=None,
                 model_path="kitsune.pkl"):

        self.AD_grace_period = AD_grace_period
        if FM_grace_period is None:
            self.FM_grace_period = AD_grace_period
        else:
            self.FM_grace_period = FM_grace_period
        self.input_precision = input_precision
        if max_autoencoder_size <= 0:
            self.m = 1
        else:
            self.m = max_autoencoder_size
        self.lr = learning_rate
        self.hr = hidden_ratio
        self.n = n
        self.normalize = normalize
        # Variables
        self.n_trained = 0  # the number of training instances so far
        self.n_executed = 0  # the number of executed instances so far
        self.v = feature_map
        self.ensembleLayer = []
        self.outputLayer = None
        self.quantize = quantize
        self.model_path = model_path
        self.norm_params_path = model_path.replace(".pkl", "_norm_params.pkl")
        if self.v is None:
            pass
            print("Feature-Mapper: train-mode, Anomaly-Detector: off-mode")
        else:
            self.__createAD__()
            print("Feature-Mapper: execute-mode, Anomaly-Detector: train-mode")
        # incremental feature clustering for the feature mapping process
        self.FM = corClust(self.n)

    def process(self, x):
        """
        If FM_grace_period+AM_grace_period has passed, then this function executes KitNET on x.
        Otherwise, this function learns from x.
        x: a np array of length n
        Note: KitNET automatically performs 0-1 normalization on all attributes.
        """
        # if all -1 it means the packet was ignored
        if x.all() == -1:
            return 0.

        if self.n_trained > self.FM_grace_period + self.AD_grace_period:  # If both the FM and AD are in execute-mode
            return self.execute(x)
        else:
            self.train(x)
            return 0.0
 
    def decision_function(self, x):
        """
        alias so it is compatible with sklearn models, input and output are all 2d arrays
        """
        anom_score = self.process(x[0])
        return [-anom_score]

    def predict(self, x):
        """
        alias for execute for it is compatible with tf models, processes in batches
        """
        return np.array([self.process(x[i]) for i in range(np.array(x).shape[0])])

    def train(self, x):
        """
    Trains the model on instance 'x'. Updates the correlation matrix during the grace period, 
    trains the ensemble and output layers otherwise. Also saves the min and max norms to a file.

    Parameters:
    x (numpy array): The instance for training.

    Returns:
    None
    """
        # If the FM is in train-mode, and the user has not supplied a feature mapping
        if self.n_trained <= self.FM_grace_period and self.v is None:
            # update the incremetnal correlation matrix
            self.FM.update(x)
            if self.n_trained == self.FM_grace_period:  # If the feature mapping should be instantiated
                self.v = self.FM.cluster(self.m)
                self.__createAD__()
                print("The Feature-Mapper found a mapping: "+str(self.n)+ \
                      " features to "+str(len(self.v))+" autoencoders.")
                print("Feature-Mapper: execute-mode, Anomaly-Detector: train-mode")
        else:  # train
            # Ensemble Layer
            S_l1 = np.zeros(len(self.ensembleLayer))
            for a in range(len(self.ensembleLayer)):
                # make sub instance for autoencoder 'a'
                xi = x[self.v[a]]
                S_l1[a] = self.ensembleLayer[a].train(xi)
            # OutputLayer
            rmse = self.outputLayer.train(S_l1)

            """save ensemble and output layer norms"""
            norm_params = {}
            for a in range(len(self.ensembleLayer)):
                norm_params[f"norm_min_{self.v[a][0]}"] = self.ensembleLayer[a].norm_min
                norm_params[f"norm_max_{self.v[a][0]}"] = self.ensembleLayer[a].norm_max
            
            norm_params["norm_min_output"] = self.outputLayer.norm_min
            norm_params["norm_max_output"] = self.outputLayer.norm_max

            with open(self.norm_params_path, 'wb') as f:
                pickle.dump(norm_params, f)

            if self.n_trained == self.AD_grace_period + self.FM_grace_period:
                pass
                print("Feature-Mapper: execute-mode, Anomaly-Detector: execute-mode")
        self.n_trained += 1

    def execute(self, x):
        if self.v is None:
            raise RuntimeError(
                'KitNET Cannot execute x, because a feature mapping has not yet been learned \
                 or provided. Try running process(x) instead.')
        else:
            self.n_executed += 1
            # Ensemble Layer
            S_l1 = np.zeros(len(self.ensembleLayer))
            for a in range(len(self.ensembleLayer)):
                # make sub inst
                xi = x[self.v[a]]
                S_l1[a] = self.ensembleLayer[a].execute(xi)
            # OutputLayer
            return self.outputLayer.execute(S_l1)

    def __createAD__(self):
        # construct ensemble layer
        for map in self.v:
            params = dA_params(n_visible=len(map), n_hidden=0, lr=self.lr, corruption_level=0, gracePeriod=0,
                                  hiddenRatio=self.hr, normalize=self.normalize,
                                  input_precision=self.input_precision, quantize=self.quantize)
            self.ensembleLayer.append(dA(params))

        # construct output layer
        params = dA_params(len(self.v), n_hidden=0, lr=self.lr, corruption_level=0, gracePeriod=0, hiddenRatio=self.hr,
                              normalize=self.normalize, quantize=self.quantize, input_precision=self.input_precision)
        self.outputLayer = dA(params)

    def get_params(self):
        return_dict = {"ensemble": []}
        for i in range(len(self.ensembleLayer)):
            return_dict["ensemble"].append(self.ensembleLayer[i].get_params())
        return_dict["output"] = self.outputLayer.get_params()
        return return_dict

    def set_params(self, new_param):
        for i in range(len(new_param["ensemble"])):
            self.ensembleLayer[i].set_params(new_param["ensemble"][i])
        self.outputLayer.set_params(new_param["output"])

    def get_torch_model(self):
        weights = self.get_params()
        model = _TorchKitNET(weights["ensemble"], weights["output"], self.v, self.n)
        path = self.model_path.replace(".pkl", ".pth")
        torch.save(model.state_dict(), path)

def squeeze_features(fv, precision):
    """rounds features to siginificant figures

    Args:
        fv (array): feature vector.
        precision (int): number of precisions to use.

    Returns:
        array: rounded array of floats.

    """

    return np.around(fv, decimals=precision)

def quantize(x, k):
    n = 2**k - 1
    return np.round(np.multiply(n, x))/n

def quantize_weights(w, k):
    x = np.tanh(w)
    q = x / np.max(np.abs(x)) * 0.5 + 0.5
    return 2 * quantize(q, k) - 1


class dA_params:
    def __init__(self, n_visible=5,
                 n_hidden=3,
                 lr=0.001,
                 corruption_level=0.0,
                 gracePeriod=10000,
                 hiddenRatio=None,
                 normalize=True,
                 input_precision=None,
                 quantize=None):

        self.n_visible = n_visible  # num of units in visible (input) layer
        self.n_hidden = n_hidden  # num of units in hidden layer
        self.lr = lr
        self.corruption_level = corruption_level
        self.gracePeriod = gracePeriod
        self.hiddenRatio = hiddenRatio
        self.normalize = normalize
        self.quantize=quantize
        self.input_precision=input_precision
        if quantize:
            self.q_wbit,self.q_abit=quantize


class dA:
    def __init__(self, params):
        self.params = params

        if self.params.hiddenRatio is not None:
            self.params.n_hidden = int(np.ceil(
                self.params.n_visible * self.params.hiddenRatio))

        # for 0-1 normlaization
        self.norm_max = np.ones((self.params.n_visible,)) * -np.Inf
        self.norm_min = np.ones((self.params.n_visible,)) * np.Inf
        self.n = 0

        self.rng = np.random.RandomState(1234)

        a = 1. / self.params.n_visible
        self.W = np.array(self.rng.uniform(  # initialize W uniformly
            low=-a,
            high=a,
            size=(self.params.n_visible, self.params.n_hidden)))

        #quantize weights
        if self.params.quantize:
            self.W=quantize_weights(self.W, self.params.q_wbit)

        self.hbias = np.zeros(self.params.n_hidden)  # initialize h bias 0
        self.vbias = np.zeros(self.params.n_visible)  # initialize v bias 0
        # self.W_prime = self.W.T

    def get_corrupted_input(self, input, corruption_level):
        assert corruption_level < 1

        return self.rng.binomial(size=input.shape,
                                 n=1,
                                 p=1 - corruption_level) * input

    # Encode
    def get_hidden_values(self, input):
        return sigmoid(np.dot(input, self.W) + self.hbias)

    # Decode
    def get_reconstructed_input(self, hidden):
        return sigmoid(np.dot(hidden, self.W.T) + self.vbias)

    def train(self, x):
        self.n = self.n + 1

        if self.params.normalize:
            # update norms
            self.norm_max[x > self.norm_max] = x[x > self.norm_max]
            self.norm_min[x < self.norm_min] = x[x < self.norm_min]

            # 0-1 normalize
            x = (x - self.norm_min) / (self.norm_max -
                                       self.norm_min + 0.0000000000000001)

        if self.params.input_precision:
            x=squeeze_features(x,self.params.input_precision)

        if self.params.corruption_level > 0.0:
            tilde_x = self.get_corrupted_input(x, self.params.corruption_level)
        else:
            tilde_x = x

        y = self.get_hidden_values(tilde_x)
        if self.params.quantize:
            y=quantize(y, self.params.q_abit)

        z = self.get_reconstructed_input(y)

        L_h2 = x - z
        L_h1 = np.dot(L_h2, self.W) * y * (1 - y)

        L_vbias = L_h2
        L_hbias = L_h1
        L_W = np.outer(tilde_x.T, L_h1) + np.outer(L_h2.T, y)

        self.W += self.params.lr * L_W
        self.hbias += self.params.lr * L_hbias
        self.vbias += self.params.lr * L_vbias

        if self.params.quantize:
            self.W=quantize_weights(self.W, self.params.q_wbit)
            self.hbias=quantize_weights(self.hbias, self.params.q_wbit)
            self.vbias=quantize_weights(self.vbias, self.params.q_wbit)

        return np.sqrt(np.mean(L_h2**2))

    def reconstruct(self, x):
        y = self.get_hidden_values(x)

        try:
            if self.params.quantize:
                y=quantize(y, self.params.q_abit)
        except AttributeError as e:
            pass
            
        z = self.get_reconstructed_input(y)
        return z

    def get_params(self):
        params={
        "W":self.W,
        "hbias":self.hbias,
        "vbias":self.vbias
        }
        return params

    def set_params(self, new_param):
        self.W=new_param["W"]
        self.hbias=new_param["hbias"]
        self.vbias=new_param["vbias"]

    def execute(self, x):
        if self.n < self.params.gracePeriod:
            return 0.0
        else:
            # 0-1 normalize
            try:
                if self.params.normalize:
                    x = (x - self.norm_min) / (self.norm_max -
                                               self.norm_min + 0.0000000000000001)

                if self.params.input_precision:
                    x=squeeze_features(x,self.params.input_precision)

            except AttributeError as e:
                pass

            z = self.reconstruct(x)
            rmse = np.sqrt(((x - z) ** 2).mean())  # MSE
            return rmse

    def inGrace(self):
        return self.n < self.params.gracePeriod

def pdf(x,mu,sigma): #normal distribution pdf
    x = (x-mu)/sigma
    return np.exp(-x**2/2)/(np.sqrt(2*np.pi)*sigma)

def invLogCDF(x,mu,sigma): #normal distribution cdf
    x = (x - mu) / sigma
    return norm.logcdf(-x) #note: we mutiple by -1 after normalization to better get the 1-cdf

def sigmoid(x):
    return 1. / (1 + np.exp(-x))

def dsigmoid(x):
    return x * (1. - x)

def tanh(x):
    return np.tanh(x)

def dtanh(x):
    return 1. - x * x

def softmax(x):
    e = np.exp(x - np.max(x))  # prevent overflow
    if e.ndim == 1:
        return e / np.sum(e, axis=0)
    else:  
        return e / np.array([np.sum(e, axis=1)]).T  # ndim = 2

def ReLU(x):
    return x * (x > 0)

def dReLU(x):
    return 1. * (x > 0)

class rollmean:
    def __init__(self,k):
        self.winsize = k
        self.window = np.zeros(self.winsize)
        self.pointer = 0

    def apply(self,newval):
        self.window[self.pointer]=newval
        self.pointer = (self.pointer+1) % self.winsize
        return np.mean(self.window)


class corClust:
    """
        A helper class for KitNET which performs a correlation-based incremental 
        clustering of the dimensions in X

        # n: the number of dimensions in the dataset
        # For more information and citation, please see our NDSS'18 paper:
        Kitsune: An Ensemble of Autoencoders for Online Network Intrusion Detection
    """
    def __init__(self,n):
        #parameter:
        self.n = n
        #varaibles
        self.c = np.zeros(n) #linear num of features
        self.c_r = np.zeros(n) #linear sum of feature residules
        self.c_rs = np.zeros(n) #linear sum of feature residules
        self.C = np.zeros((n,n)) #partial correlation matrix
        self.N = 0 #number of updates performed

    # x: a numpy vector of length n
    def update(self,x):
        self.N += 1
        self.c += x
        c_rt = x - self.c/self.N
        self.c_r += c_rt
        self.c_rs += c_rt**2
        self.C += np.outer(c_rt,c_rt)

    # creates the current correlation distance matrix between the features
    def corrDist(self):
        c_rs_sqrt = np.sqrt(self.c_rs)
        C_rs_sqrt = np.outer(c_rs_sqrt,c_rs_sqrt)
        C_rs_sqrt[C_rs_sqrt==0] = 1e-100 #this protects against dive by zero erros (occurs when a feature is a constant)
        D = 1-self.C/C_rs_sqrt #the correlation distance matrix
        D[D<0] = 0 #small negatives may appear due to the incremental fashion in which we update the mean. Therefore, we 'fix' them
        return D

    # clusters the features together, having no more than maxClust features per cluster
    def cluster(self,maxClust):
        D = self.corrDist()
        Z = linkage(D[np.triu_indices(self.n, 1)])  # create a linkage matrix based on the distance matrix
        if maxClust < 1:
            maxClust = 1
        if maxClust > self.n:
            maxClust = self.n
        map = self.__breakClust__(to_tree(Z),maxClust)
        return map

    # a recursive helper function which breaks down the dendrogram branches until all clusters have no more than maxClust elements
    def __breakClust__(self,dendro,maxClust):
        if dendro.count <= maxClust: #base case: we found a minimal cluster, so mark it
            return [dendro.pre_order()] #return the origional ids of the features in this cluster
        return self.__breakClust__(dendro.get_left(),maxClust) + self.__breakClust__(dendro.get_right(),maxClust)



import time
from matplotlib import ticker
import torch
import random

import numpy as np
import torch.nn as nn

from torch import Tensor
from typing import Any
from scapy.all import *


import titli.fe.corClust as CC
from titli.fe import AfterImage, NetStat

import matplotlib.pyplot as plt

class RMSELoss(nn.Module):
    def __init__(self):
        super().__init__()
        self.mse = nn.MSELoss()

    def forward(self, yhat, y):
        return torch.sqrt(self.mse(yhat, y))

class BaseAutoencoder(nn.Module):
    def __init__(self, input_size, hidden_size):
        super().__init__()
        self.norm_max = np.ones((input_size,)) * -np.Inf
        self.norm_min = np.ones((input_size,)) * np.Inf
        self.encoder = nn.Linear(input_size, hidden_size)
        self.decoder = nn.Linear(hidden_size, input_size)

    def forward(self, x):
        x = torch.sigmoid(self.encoder(x))
        x = torch.sigmoid(self.decoder(x))
        return x
    
    def normalise(self, x: np.ndarray, train_mode=False) -> np.ndarray:
        if train_mode:
            self.norm_max[x > self.norm_max] = x[x > self.norm_max]
            self.norm_min[x < self.norm_min] = x[x < self.norm_min]

        # 0-1 normalize
        x = (x - self.norm_min) / (self.norm_max -
                                    self.norm_min + 0.0000000000000001)
        return x


class PyTorchKitsune(nn.Module):

    def __init__(self, num_features: int, **kwargs):
        self.state = NetStat()
        self.fe = AfterImage(state=self.state)
        self.hidden_rate = 0.75
        self.FMgrace_rate = 0.2
        self.train_mode = False
        self.num_features = num_features
        super().__init__(**kwargs)
        self.FM = CC.corClust(self.num_features)

    def _define_model(self) -> nn.Module:

        self.tails = nn.ModuleList(
            [BaseAutoencoder(len(c), int(np.ceil(len(c) * self.hidden_rate))) for c in self.clusters]
        )

        for tail in self.tails:
            input_size = len(tail.norm_max)
            a = 1. / input_size

            # Initialize the weights uniformly between -a and a
            nn.init.uniform_(tail.encoder.weight, -a, a)
            
            # Set the decoder weights as the transpose of the encoder weights
            tail.decoder.weight.data = tail.encoder.weight.data.t()

            # Initialize the biases to 0
            nn.init.zeros_(tail.encoder.bias)
            nn.init.zeros_(tail.decoder.bias)
        
        self.head = BaseAutoencoder(
            len(self.clusters), int(np.ceil(len(self.clusters) * self.hidden_rate))
        )
        # Get the input size for the head
        input_size = len(self.head.norm_max)
        a = 1. / input_size

        # Initialize the weights of the head uniformly between -a and a
        nn.init.uniform_(self.head.encoder.weight, -a, a) 
        
        # Set the decoder weights as the transpose of the encoder weights
        self.head.decoder.weight.data = self.head.encoder.weight.data.t()
        nn.init.zeros_(self.head.encoder.bias)
        nn.init.zeros_(self.head.decoder.bias)

    def normalise(self, x: Tensor, train_mode=False) -> Tensor:
        # make a copy of x and convert to numpy
        x_alias = x.clone().detach().cpu().numpy()
        if train_mode:
            self.norm_max[x_alias > self.norm_max] = x_alias[x_alias > self.norm_max]
            self.norm_min[x_alias < self.norm_min] = x_alias[x_alias < self.norm_min]

        # 0-1 normalize
        x = (x - torch.tensor(self.norm_min)) / (torch.tensor(self.norm_max) -
                                    torch.tensor(self.norm_min) + 0.0000000000000001)
        return x
    
    def forward(self, x):
        x = x.reshape(-1, self.num_features)

        x_clusters = []
        for c in self.clusters:
            x_cluster = x[:, c]
            x_clusters.append(x_cluster)

        tail_losses = []
        for tail, c in zip(self.tails, x_clusters):
            c = tail.normalise(c.flatten(), train_mode=self.train_mode)
            # reshape convert to tensor, and pass to device
            c = torch.tensor(c).unsqueeze(0).float().to(self.device) # TODO: This has to be tensor from the start!!!!!!!!!!!!!!!
            output = tail(c)
            if self.train_mode: # TODO: Confirm that this is not Log in training mode!!!!!!!!!!!!
                # loss = torch.log(self.rmse(output, c))
                loss = self.rmse(output, c)
            else:
                loss = self.rmse(output, c)
            if loss.data == 0:
                loss.data = torch.tensor(1e-2)
            tail_losses.append(loss)

        tails = torch.stack(tail_losses)
        tails = self.normalise(tails, train_mode=self.train_mode).float()
        x_hat = self.head(tails)

        return x_hat, tails
    
    def train(self, pcap_path: str) -> None:
        self.train_rmse = []
        self.print_interval = 100
        self.train_mode = True
        self.device = torch.device("mps" if torch.cuda.is_available() else "cpu")
        self.rmse = RMSELoss()
        self.FMgrace = np.floor(self.FMgrace_rate * len(rdpcap(pcap_path)))
        self.dataset = pcap_path.split("/")[-1].split(".")[0]
        
        print("Training Feature Mapping (FM) phase")
        for i, packet in enumerate(PcapReader(pcap_path)):

            # get feature vector
            traffic_vector = self.fe.get_traffic_vector(packet)

            # extract features
            x = self.fe.update(traffic_vector)

            # check for FMgrace period
            if i < self.FMgrace:
                self.FM.update(x)
                continue

            if i == self.FMgrace:
                self.clusters = self.FM.cluster(maxClust=10)
                self._define_model()
                optimiser = torch.optim.Adam(self.parameters(), lr=0.0001)
                self.norm_max = np.ones((len(self.clusters),)) * -np.Inf
                self.norm_min = np.ones((len(self.clusters),)) * np.Inf
                print("Training Autoencoder phase")
                continue

            # forward pass
            x_hat, tails = self(x)

            # loss
            loss = self.rmse(x_hat, tails)

            self.train_rmse.append(loss.data)

            optimiser.zero_grad()
            loss.backward()
            optimiser.step()

            # # Making the decoder weight the transpose of the encoder weight
            # for tail in self.tails:
            #     tail.decoder.weight.data = tail.encoder.weight.data.t()
            # self.head.decoder.weight.data = self.head.encoder.weight.data.t()

            # Making the encoder weight the transpose of the decoder weight
            for tail in self.tails:
                tail.encoder.weight.data = tail.decoder.weight.data.t()
            self.head.encoder.weight.data = self.head.decoder.weight.data.t()

            if (i + 1) % self.print_interval == 0:
                print(f"Packet: {i} | Loss: {loss.data}")
        
        self.threshold = self.get_threshold(pcap_path)

        return self.threshold
    
    def get_threshold(self, pcap_path: str) -> float:
        print("Calculating threshold")
        self.threshold_rmse = []
        self.print_interval = 10
        self.train_mode = False
        self.device = torch.device("mps" if torch.cuda.is_available() else "cpu")
        self.rmse = RMSELoss()
        
        for _, packet in enumerate(PcapReader(pcap_path)):
            
            # get feature vector
            traffic_vector = self.fe.get_traffic_vector(packet)

            # extract features
            x = self.fe.update(traffic_vector)

            # forward pass
            x_hat, tails = self(x)

            # loss
            loss = self.rmse(x_hat, tails)
            if loss.data == 0:
                loss.data = torch.tensor(1e-2)
            self.threshold_rmse.append(loss.data)

        log_re = np.log(self.threshold_rmse)
        mean = np.mean(log_re)
        std = np.std(log_re)
        threshold_std = np.exp(mean + 3 * std)
        threshold_max = max(self.threshold_rmse)
        threshold = min(threshold_max, threshold_std)

        return threshold

    def infer(self, pcap_path: str) -> None:
        print("Inferencing")
        self.rmse_array = []
        self.print_interval = 10
        self.train_mode = False
        self.device = torch.device("mps" if torch.cuda.is_available() else "cpu")
        self.rmse = RMSELoss()
        
        for i, packet in enumerate(PcapReader(pcap_path)):
            
            # get feature vector
            traffic_vector = self.fe.get_traffic_vector(packet)

            # extract features
            x = self.fe.update(traffic_vector)

            # forward pass
            x_hat, tails = self(x)

            # loss
            loss = self.rmse(x_hat, tails)
            if loss.data == 0:
                loss.data = torch.tensor(1e-2)
            self.rmse_array.append(loss.data)

            if (i + 1) % self.print_interval == 0:
                print(f"Inferencing: {i} | Loss: {loss.data}")

    def plot_kitsune(self, threshold, out_image = "kitsune_plot.png"):
        _ = plt.get_cmap('Set3')

        f, ax1 = plt.subplots(constrained_layout=True, figsize=(10, 5), dpi=600)
        x_val = np.arange(len(self.rmse_array))

        try:
            ax1.scatter(x_val, self.rmse_array, s=1, c='#00008B')
        except:
            ax1.scatter(x_val, self.rmse_array, s=1, alpha=1.0, c='#FF8C00')

        ax1.axhline(y=threshold, color='b', linestyle='-')
        ax1.set_yscale("log")
        ax1.set_title(f"Anomaly Scores of Kitsune Execution Phase: {self.dataset}")
        ax1.set_ylabel("RMSE")
        ax1.set_xlabel("Packet Index")

        f.savefig(out_image)
        print("plot path:", out_image)
        plt.close()
